
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtbucket.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"

static char vcid[] = "$Id: vdtbucket.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtbucket) 
 
typedef struct vdt_bucket_t {
  LIST  ptl;
}              vdt_bucket_t;

typedef struct vdt_bucket_array_t {
  vdt_bucket_t *za;
}              vdt_bucket_array_t;

typedef struct vdt_bucket_array_of_array_t {
  vdt_bucket_array_t *ya;
}              vdt_bucket_array_of_array_t;

typedef struct vdt_buckets_t {
  CkitPoint3                   origin;
  double                       bucket_size;
  int                          nx, ny, nz;
  vdt_bucket_array_of_array_t *xa;
  int                          tpt;
  int                          npt;
  PTP                         *pta;
  int                          find_max_num_objects;
}              vdt_buckets_t;

vdt_buckets
vdt_bucket_new (G3D_box_t *box, double bucket_size)
{
  vdt_buckets_t *b;
  CkitPoint3 v;
  int i, j, k;
  
  b = NEWP_OBJ (vdt_buckets_t); CHECK_MEM (b);

  if (bucket_size <= 0) {
    bucket_size = (  G3D_BOX_RANGE ((*box), x)
                   + G3D_BOX_RANGE ((*box), y)
                   + G3D_BOX_RANGE ((*box), z)) / 3 / 10;
  }
  b->bucket_size = bucket_size;
  b->origin = box->lower;
  VECMAC_SCALAR (b->origin, -=, 0.5 * bucket_size);
  VECMAC_3_VEC (v, =, box->upper, -, b->origin);
  VECMAC_SCALAR (v, +=, 0.5 * bucket_size);
  b->nx = (int) (v.x / b->bucket_size) + 1;
  if (b->origin.x + b->nx * b->bucket_size < box->upper.x)
    errexit ("bad");
  b->ny = (int) (v.y / b->bucket_size) + 1;
  if (b->origin.y + b->ny * b->bucket_size < box->upper.y)
    errexit ("bad");
  b->nz = (int) (v.z / b->bucket_size) + 1;
  if (b->origin.z + b->nz * b->bucket_size < box->upper.z)
    errexit ("bad");

  b->xa = NEWP_OBJ_ARRAY (b->nx, vdt_bucket_array_of_array_t);
  CHECK_MEM (b->xa);
  FOR (i, b->nx) {
    b->xa[i].ya = NEWP_OBJ_ARRAY (b->ny, vdt_bucket_array_t);
    CHECK_MEM (b->xa[i].ya);
    FOR (j, b->ny) {
      b->xa[i].ya[j].za = NEWP_OBJ_ARRAY (b->nz, vdt_bucket_t);
      CHECK_MEM (b->xa[i].ya[j].za);
      FOR (k, b->nz) {
        b->xa[i].ya[j].za[k].ptl = NULL;
      }
    }
  }

  b->tpt = 100;
  b->npt = 0;
  b->pta = NEWP_OBJ_ARRAY (b->tpt, PTP); CHECK_MEM (b->pta);
  b->find_max_num_objects = INT_MAX;
    
  return b;
}

   
#define BSTORE(b, PT)                                                                   \
  {                                                                                     \
    if (b->npt >= b->tpt) {                                                             \
      b->tpt *= 2;                                                                      \
      b->pta = NEWP_REALLOC_OBJ_ARRAY (b->pta, b->tpt, PTP); CHECK_MEM (b->pta);        \
    }                                                                                   \
    b->pta[b->npt] = PT; b->npt++;                                                      \
  }

int
vdt_bucket_search_box (vdt_buckets b, G3D_box_t *box, int max_objects)
{
  int xs, ys, zs, xe, ye, ze;
  int i, j, k;
  PTP pt;
  LIST l;

  if (max_objects <= 0) b->find_max_num_objects = INT_MAX;
  else                  b->find_max_num_objects = max_objects;
  
  xs = (int) ((box->lower.x - b->origin.x) / b->bucket_size);
  if (xs < 0) xs = 0;
  ys = (int) ((box->lower.y - b->origin.y) / b->bucket_size);
  if (ys < 0) ys = 0;
  zs = (int) ((box->lower.z - b->origin.z) / b->bucket_size);
  if (zs < 0) zs = 0;
  
  xe = (int) ((box->upper.x - b->origin.x) / b->bucket_size);
  if (xe >= b->nx) xe = b->nx - 1;
  ye = (int) ((box->upper.y - b->origin.y) / b->bucket_size);
  if (ye >= b->ny) ye = b->ny - 1;
  ze = (int) ((box->upper.z - b->origin.z) / b->bucket_size);
  if (ze >= b->nz) ze = b->nz - 1;

  b->npt = 0;
  for (i = xs; i <= xe; i++) {
    for (j = ys; j <= ye; j++) {
      for (k = zs; k <= ze; k++) {
        if (b->xa[i].ya[j].za[k].ptl) {
          l  = b->xa[i].ya[j].za[k].ptl;
          pt = NULL;
          while ((pt = (PTP) get_list_next (l, pt))) {
            if (G3D_PT_INSIDE_BOX ((*box), pt->loc)) {
              BSTORE (b, pt);
              if (b->find_max_num_objects == b->npt) return b->npt;
            }
          }
        }
      }
    }
  }

  return b->npt;
}

int
vdt_bucket_add_pt (vdt_buckets b, PTP pt)
{
  int i, j, k;
  LIST l;
  
  i = (int) ((pt->loc.x - b->origin.x) / b->bucket_size);
  j = (int) ((pt->loc.y - b->origin.y) / b->bucket_size);
  k = (int) ((pt->loc.z - b->origin.z) / b->bucket_size);
  assert (i >= 0 && i < b->nx);
  assert (j >= 0 && j < b->ny);
  assert (k >= 0 && k < b->nz);
  l = b->xa[i].ya[j].za[k].ptl;
  if (!l) {
    l = make_list ();
    CHECK_MEM (l);
    b->xa[i].ya[j].za[k].ptl = l;
  }
  add_to_tail (l, pt);
  return TRUE;
}

int
vdt_bucket_num_of_points_found (vdt_buckets b)
{
  return b->npt;
}

PTP
vdt_bucket_pt_at (vdt_buckets b, int i)
{
  return b->pta[i];
}

void
vdt_bucket_free (vdt_buckets b) 
{
  int i, j, k;
  LIST l;
  FOR (i, b->nx) {
    FOR (j, b->ny) {
      FOR (k, b->nz) {
        if (b->xa[i].ya[j].za[k].ptl) {
          l  = b->xa[i].ya[j].za[k].ptl;
          free_list (l, NOT_NODES);
        }
      }
      NEWP_FREE_OBJ (b->xa[i].ya[j].za);
    }
    NEWP_FREE_OBJ (b->xa[i].ya);
  }
  NEWP_FREE_OBJ (b->xa);
  NEWP_FREE_OBJ (b->pta);
  NEWP_FREE_OBJ (b);
}
