
/* 
      Variational Delaunay Triangulation
             (C) 1998-2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtlp.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"

static char vcid[] = "$Id: vdtlp.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtlp) 
 



/*
   Create a linear programming solver context.

   Suggestion for improvement:  Try setting upper bounds instead of using
   constraints.
 */
lprec *
vdt_lp_make (CkitVector3 input_point[],
             int num_input_points,
             CkitVector3 *sampling_point)
{
  int i;
  CkitVector3 v;
  int num_constraints;
  lprec *lp = NULL;
  static REAL *row = NULL;

  num_constraints =
    (  1                        /* sum of lambdas */
     + num_input_points         /* lambdas <= 1 */
     + 3                        /* reproduction of the sampling point */
     + 1                        /* indices start from one */
     );
    
  lp = make_lp (0, num_input_points);

  row = NEWP_REALLOC_OBJ_ARRAY (row, num_constraints, double);

  /* objective function */
  for (i = 1; i <= num_input_points; i++) {
    VECMAC_3_VEC (v, =, input_point[i-1], -, (*sampling_point));
    row[i] = VECMAC_DOT (v, v);
  }
  set_obj_fn (lp, row);

  /* sum of lambdas == 1 */
  for (i = 1; i <= num_input_points; i++) {
    row[i] = 1;
  }
  add_constraint (lp, row, EQ, 1.0);

  /* constrain lambdas to be less or equal to one */
  for (i = 0; i <= num_input_points; i++) { row[i] = 0; }
  for (i = 1; i <= num_input_points; i++) {
    if (i > 1) {
      row[i-1] = 0;
    }
    row[i] = 1;
    add_constraint (lp, row, LE, 1.0);
  }

  /* reproduce the sampling point */
  for (i = 1; i <= num_input_points; i++) {
    row[i] = input_point[i-1].x;
  }
  add_constraint (lp, row, EQ, sampling_point->x);
  for (i = 1; i <= num_input_points; i++) {
    row[i] = input_point[i-1].y;
  }
  add_constraint (lp, row, EQ, sampling_point->y);
  for (i = 1; i <= num_input_points; i++) {
    row[i] = input_point[i-1].z;
  }
  add_constraint (lp, row, EQ, sampling_point->z);
  
  set_minim (lp);

  return lp;
}

/*
   Print the definition of the linear programming problem
   in lp_solve readable form.
 */
int
vdt_lp_setup (CkitVector3 input_point[],
              int num_input_points,
              CkitVector3 *sampling_point, FILE *fp)
{
  int i, nip;
  CkitVector3 v;
  int constraint;
  double cval;

  nip = num_input_points;

  fprintf (fp, "min: ");
  for (i = 0; i < nip; i++) {
    VECMAC_3_VEC (v, =, input_point[i], -, (*sampling_point));
    fprintf (fp, "%g x%d", VECMAC_DOT (v, v), i+1);
    if (i + 1 < nip) fprintf (fp, " + ");
  }
  fprintf (fp, ";\n");

  constraint = 1;
  
  fprintf (fp, "c%d: ", constraint++);
  for (i = 0; i < nip; i++) {
    fprintf (fp, "x%d", i+1);
    if (i + 1 < nip) fprintf (fp, " + ");
  }
  fprintf (fp, " = 1;\n");

  for (i = 0; i < nip; i++) {
    fprintf (fp, "c%d: x%d <= 1;\n", constraint++, i+1);
  }

  fprintf (fp, "c%d: ", constraint++);
  for (i = 0; i < nip; i++) {
    cval = input_point[i].x;
    if (cval != 0) {
      fprintf (fp, "%g x%d", cval, i+1);
      if (i + 1 < nip) fprintf (fp, " + ");
    }
  }
  fprintf (fp, " = %g;\n", sampling_point->x);

  fprintf (fp, "c%d: ", constraint++);
  for (i = 0; i < nip; i++) {
    cval = input_point[i].y;
    if (cval != 0) {
      fprintf (fp, "%g x%d", cval, i+1);
      if (i + 1 < nip) fprintf (fp, " + ");
    }
  }
  fprintf (fp, " = %g;\n", sampling_point->y);

  fprintf (fp, "c%d: ", constraint++);
  for (i = 0; i < nip; i++) {
    cval = input_point[i].z;
    if (cval != 0) {
      fprintf (fp, "%g x%d", input_point[i].z, i+1);
      if (i + 1 < nip) fprintf (fp, " + ");
    }
  }
  fprintf (fp, " = %g;\n", sampling_point->z);

  return 0;
}
