
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtpt.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $
   */

#include "vdtmaster.h"


static char vcid[] = "$Id: vdtpt.c,v 1.1.1.1 2001/07/24 02:11:23 pkrysl Exp $";
DEF_VDT_MODULE_ID (vdtpt) 
 

/*
   Add a point to the triangulation.
   A point is added to the triangulation.
 */
vdt_pt_t *
vdt_pt_new (void *id,
            CkitPoint3 *loc,  /* location in 3-D space */
            double mesh_size    /* mesh size */
            )
{
  vdt_pt_t *pt;
   
  pt = NEWP_OBJ (vdt_pt_t); CHECK_MEM (pt);
  pt->id                 = id;
  pt->origloc = pt->loc  = *loc;
  pt->ffl                = NULL;
  pt->bsfl               = NULL;
  pt->mesh_size          = mesh_size;
#if GIFACE
  pt->go                 = NULL;
  pt->atgo               = NULL;
#endif
  pt->num_conn_tets      = 0;

  return pt;
}



#if GIFACE
/*
   Draw a point
 */
void
vdt_pt_generate_graphics (vdt_pt_t *pt, vdt_t *vdt)
{
  WCRec pts[2];
  if (vdt->gmodel) {
    if (!pt->go) {
      char buf[32];
      pts[0] = pt->loc;
      pt->go = CreateMarker3D (pts);
      EMNoDrawAddGraphicsToModel (vdt->gmodel, pt->go);
      EASValsSetMType (FILLED_CIRCLE_MARKER);
      EASValsSetMSize (4);
      if (pt->id > vdt->input_max_pt_id) {
        EASValsSetColor (vdt_color_to_pixel ("orange"));
        EASValsSetLayer (VDT_ADD_PT_LAYER);        
      } else {
        EASValsSetColor (vdt_color_to_pixel ("cyan"));
        EASValsSetLayer (VDT_PT_LAYER);
      }
      EGWithMaskChangeAttributes (COLOR_MASK | MTYPE_MASK |
                                  MSIZE_MASK | LAYER_MASK, pt->go);
      
      sprintf (buf, "_%d", pt->id);
      pts[0] = pt->loc;
      pt->atgo = CreateAnnText3D (pts, buf);
      EMNoDrawAddGraphicsToModel (vdt->gmodel, pt->atgo);
      EASValsSetColor (vdt_color_to_pixel ("magenta"));
      EASValsSetLayer (VDT_PT_NUM_LAYER);
      EGWithMaskChangeAttributes (COLOR_MASK | LAYER_MASK, pt->atgo);
    } else {
      pts[0] = pt->loc;
      EGModifyGraphicsGeometry (pt->go, pts);
      EGModifyGraphicsGeometry (pt->atgo, pts);
    }
  }
}
#endif


/*
   Generate randomly offset sampling point.
 */
int
vdt_pt_random_sampling_point (vdt_pt_t *pt,
                              double displ_magnitude, CkitPoint3 *sp)
{
  CkitPoint3 rv;
  CkitPoint3 avgn, n;
  vdt_bsf_t *bsf;

  rv.x = vdt_rand_0_to_1 (0) * displ_magnitude;
  rv.y = vdt_rand_0_to_1 (1) * displ_magnitude;
  rv.z = vdt_rand_0_to_1 (2) * displ_magnitude;

  if (pt->bsfl) { /* constrain by surface normals to be inside the region */
    VECMAC_SCALAR (avgn, =, 0);
    loop_through_list (pt->bsfl, bsf, vdt_bsf_t *) {
      vdt_bsf_get_normal (bsf, &n);
      VECMAC_2_VEC (avgn, +=, n);
    }
    if (VECMAC_NORM (avgn) <= 0) {
      warn ("Zero normal");
      return FALSE;
    } else {
      VECMAC_NORMALIZE (avgn);
      VECMAC_3_VEC ((*sp), =, pt->loc, -, displ_magnitude * avgn);
      if (VECMAC_DOT (avgn, rv) > 0) VECMAC_2_VEC (rv, =, -rv);
      VECMAC_2_VEC ((*sp), +=, displ_magnitude / 2 * rv);
    }
  }

  VECMAC_3_VEC ((*sp), =, pt->loc, +, rv);

  return TRUE;
}


/*
   Find a front face given remaining two points.
 */
vdt_ff_t *
vdt_pt_update_ffl (vdt_pt_t *pt, vdt_pt_t *op0, vdt_pt_t *op1,
                   vdt_tet_t *tet, int face, vdt_t *vdt)
{
  vdt_ff_t *ff;

  if (!pt->ffl) {
    /* new front face */
    pt->ffl = make_list (); CHECK_MEM (pt->ffl);
    ff = vdt_ff_add (tet, face, pt, op0, op1, vdt, NULL);
    add_to_tail (pt->ffl, ff);
    return ff;
  } else {
    /* does pt have the front face? if so, delete it from the list ... */
    loop_through_list (pt->ffl, ff, vdt_ff_t *) {
      if (   (ff->pt[1] == op0 && ff->pt[2] == op1)
          || (ff->pt[1] == op1 && ff->pt[2] == op0)
          ) {
        unlink_current_next (pt->ffl);
        return ff;
      }
    }
    /* ... otherwise add it to the list */
    ff = vdt_ff_add (tet, face, pt, op0, op1, vdt, NULL);
    add_to_tail (pt->ffl, ff);
    return ff; 
  }
}



void
vdt_pt_attach_bsf (vdt_pt_t *pt, vdt_bsf_t *bsf)
{
  if (!pt->bsfl) pt->bsfl = make_list ();
  add_to_tail (pt->bsfl, bsf);
}

void
vdt_pt_detach_bsf (vdt_pt_t *pt, vdt_bsf_t *bsf)
{
  if (!pt->bsfl) errexit ("no bsfl");
  delete_list_node (pt->bsfl, bsf);
}


/*
   Check if the point polygon on input is quasi-convex.
   Quasi in the sense that an angle tolerance is used to decide whether
   two sides subtending an angle close to pi are almost co-linear; if
   they are, the polygon is reported as non-convex.  The polygon is
   also reported as non-convex for identical points on input.
   The tolerance angle is assumed to be *small* (say, less than 0.1 rad).
 */
int
vdt_pt_convex_poly (vdt_pt_t *pt[], int npt, double minangle)
{
  CkitPoint3 a, b, n = {0,0,0}, prevn = {0,0,0};
  int i;
  double aa, bb, adb;
  double sminadb2 = (1 - 2 * minangle) * (1 - 2 * minangle);
  
  VECMAC_3_VEC (a, =, pt[0]->loc, -, pt[npt-1]->loc);
  aa = VECMAC_DOT (a, a);
  if (aa <= 0) return FALSE;    /* for identical points */
  VECMAC_3_VEC (b, =, pt[1]->loc, -, pt[0]->loc);
  bb = VECMAC_DOT (b, b);
  if (bb <= 0) return FALSE;    /* for identical points */
  VECMAC_CROSS (n, =, a, x, b);
  adb = VECMAC_DOT (a, b);
  if (adb * adb > sminadb2 * aa * bb) return FALSE; /* co-linear segments  */
  prevn = n;
  a     = b;
  aa    = bb;

  for (i = 1; i < npt-1; i++) {
    VECMAC_3_VEC (b, =, pt[i+1]->loc, -, pt[i]->loc);
    bb = VECMAC_DOT (b, b);
    if (bb <= 0) return FALSE;    /* for identical points */
    VECMAC_CROSS (n, =, a, x, b);
    adb = VECMAC_DOT (a, b);
    if (adb * adb  > sminadb2 * aa * bb) return FALSE; /* co-linear segments  */
    if (VECMAC_DOT (n, prevn) <= 0) return FALSE;
    prevn = n;
    a     = b;
    aa    = bb;
  }

  VECMAC_3_VEC (b, =, pt[0]->loc, -, pt[npt-1]->loc);
  bb = VECMAC_DOT (b, b);
  if (bb <= 0) return FALSE;    /* for identical points */
  VECMAC_CROSS (n, =, a, x, b);
  adb = VECMAC_DOT (a, b);
  if (adb * adb  > sminadb2 * aa * bb) return FALSE; /* co-linear segments  */
  if (VECMAC_DOT (n, prevn) <= 0) return FALSE;

  return TRUE;
}

static int
has_pt (FFP ff, PTP pt)
{
  return (pt == ff->pt[0] || pt == ff->pt[1] || pt == ff->pt[2]);
}

static int
ff_duplicated_in_list (LIST ffl, FFP ff)
{
  FFP off;
  int result = FALSE;
  LIST offl = copy_list (ffl, NOT_NODES);
  loop_through_list (offl, off, FFP) {
    if (off != ff) {
      if (has_pt (off, ff->pt[0]) &&
          has_pt (off, ff->pt[1]) &&
          has_pt (off, ff->pt[2])) {
        result = TRUE; goto exit;
      }
    }
  }
 exit:
  free_list (offl, NOT_NODES);
  return result;
}

static void
drop_duplicate_ffs (LIST ffl)
{
  FFP ff;
  loop_through_list (ffl, ff, FFP) {
    if (ff_duplicated_in_list (ffl, ff)) {
      ff->is_active = FALSE;
    }
  }
 again:
  ff = NULL;
  while (ff = (FFP) get_list_next (ffl, ff)) {
    if (!ff->is_active) {
      delete_list_node (ffl, ff);
      goto again;
    }
  }
}

void
vdt_pt_deactivate_dupl_ff (PTP pt)
{
  if (pt->ffl && list_length (pt->ffl) > 1)
    drop_duplicate_ffs (pt->ffl);
}
