
/* 
      Variational Delaunay Triangulation
             (C) 1998--2000 Petr Krysl

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   $Id: vdtpub.c,v 1.2 2004/04/23 19:11:16 cvs Exp $
   */

#include "vdtmaster.h"
#include "vdtpub.h"

static char vcid[] = "$Id: vdtpub.c,v 1.2 2004/04/23 19:11:16 cvs Exp $";
DEF_VDT_MODULE_ID (vdtpub) 
 
/*
   Create a new VDT mesher.
 */
VDT /* mesher handle or NULL if not enough memory */
VDT_new_mesher (void)
{
  return (VDT) vdt_new ();
}

static int
load_mesher (vdt_t *vdt)
{
  PTP pt;
  BSFP bsf;
  int i;

  pt = (PTP) get_list_next (vdt->pt_as_read, NULL);
  G3D_INIT_BOX (vdt->bbox, pt->loc);
  while ((pt = (PTP) get_list_next (vdt->pt_as_read, pt))) {
    G3D_UPDT_BOX (vdt->bbox, pt->loc);
  }
  
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pt_to_insert_as_if_gen, pt))) {
    G3D_UPDT_BOX (vdt->bbox, pt->loc);
  }
  G3D_INFLATE_BOX (vdt->bbox, vdt->def_mesh_size);
  
  pt = NULL;
  while ((pt = (PTP) get_list_next (vdt->pt_as_read, pt))) {
    vdt_attach_pt (vdt, pt);
  }
  flush_list (vdt->pt_as_read, NOT_NODES);

  bsf = NULL;
  while ((bsf = (BSFP) get_list_next (vdt->bsf_as_read, bsf))) {
    FOR (i, 3) {
      bsf->pt[i] = vdt_find_pt (vdt, bsf->pt[i]);
      if (!bsf->pt[i]) {
        vdt_set_last_error_msg (vdt, "Point not found");
        return FALSE;
      }
    }
    vdt_bsf_connect (bsf, bsf->pt[0], bsf->pt[1], bsf->pt[2], vdt);
    vdt_attach_bsf (vdt, bsf);
  }
  flush_list (vdt->bsf_as_read, NOT_NODES);
  return TRUE;
}

/*
   Add a point.
   A point is added to the mesher input data.
 */
int /* FALSE if (vdt==NULL), or if
       (id==0), or if there was not enough memory */
VDT_add_pt (void *id, /* unique id (>0) */
            double x, /* x-coordinate of the point */
            double y, /* y-coordinate of the point */
            double z, /* z-coordinate of the point */
            double mesh_size, /* mesh size associated with the
                                 point (can be supplied as 0; in that
                                 case its value is ignored)
                               */
            VDT vdt /* mesher handle */
            )
{
  CkitPoint3 loc;
  PTP pt;
  if (!vdt) {
    vdt_set_last_error_msg (vdt, "NULL passed for mesher");
    return FALSE;
  }
  if (id == 0) {
    vdt_set_last_error_msg (vdt, "Point id <= 0");
    return FALSE;
  }
  VECMAC_3_SCALAR (loc, =, x, y, z);
  if (mesh_size <= 0) {
    if (vdt->def_mesh_size <= 0) {
      vdt_set_last_error_msg (vdt, "Default mesh size <= 0");
      return FALSE;
    } else 
      mesh_size = vdt->def_mesh_size;
  }
  if ((pt = vdt_pt_new (id, &loc, mesh_size)) != NULL) {
    vdt_perturb_loc (vdt->bdry_pt_shift_mult, mesh_size, &pt->loc);
    add_to_tail (vdt->pt_as_read, pt);
    return TRUE;
  } else {
    vdt_set_last_error_msg (vdt, "Failed to allocate new point");
    return FALSE;
  }
}

/*
   Add a point.
   A point is added to the mesher input data, however, the point is
   first checked if it is not too close to other points or to
   the boundary surfaces.
 */
int /* FALSE if (vdt==NULL), or if
       (id==0), or if there was not enough memory */
VDT_add_pt_with_check (void *id, /* unique id (>0) */
                       double x, /* x-coordinate of the point */
                       double y, /* y-coordinate of the point */
                       double z, /* z-coordinate of the point */
                       double mesh_size, /* mesh size associated with the
                                            point (can be supplied as 0; in that
                                            case its value is ignored)
                                            */
                       VDT vdt /* mesher handle */
                       )
{
  CkitPoint3 loc;
  PTP pt;
  if (!vdt) {
    vdt_set_last_error_msg (vdt, "NULL passed for mesher");
    return FALSE;
  }
  if (id == 0) {
    vdt_set_last_error_msg (vdt, "Point id <= 0");
    return FALSE;
  }
  VECMAC_3_SCALAR (loc, =, x, y, z);
  if (mesh_size <= 0) {
    if (vdt->def_mesh_size <= 0) {
      vdt_set_last_error_msg (vdt, "Default mesh size <= 0");
      return FALSE;
    } else 
      mesh_size = vdt->def_mesh_size;
  }
  if ((pt = vdt_pt_new (id, &loc, mesh_size)) != NULL) {
    vdt_perturb_loc (vdt->bdry_pt_shift_mult, mesh_size, &pt->loc);
    add_to_tail (vdt->pt_to_insert_as_if_gen, pt);
    return TRUE;
  } else {
    vdt_set_last_error_msg (vdt, "Failed to allocate new point");
    return FALSE;
  }
}

/*
   Add a boundary facet.
   A boundary facet is added to the mesher input data.
   The facet is assumed to be oriented so that its points
   are traversed counterclockwise when looking against the normal.
   The normal points outside (in other words, it is the outer normal).
 */
int /* FALSE if inconsistent data have been supplied
       (points not found, or (surf_id==0),
       or (reg_id[0]==0), or if (vdt==NULL)),
       or if there was not enough memory
     */
VDT_add_bsf (void *ptid[3], /* identifiers of points; the points
                             need to be defined first by vdt_add_pt().
                             The order in which points are defined
                             determines the normal so that the points
                             are counterclockwise when looking
                             against the normal.  The normal must be
                             either the outer normal (for the outer
                             surfaces), or can be set arbitrarily
                             (for inner, interface surfaces).
                           */
             void *surf_id, /* identifier of the surface the
                             facet is on; positive number
                           */
             void *reg_id[2], /* identifiers of regions on each side
                              of the facet (pointers or positive integers) if there
                              is a region on the given side of the facet;
                              NULL otherwise;
                              Note that reg_id[1] is on the
                              side to which the normal points,
                              reg_id[0] is on the other side.
                              If the facet is incident on one region
                              only, it must be the zeroth element of reg_id
                            */
             VDT vdt /* mesher handle */
             )
{
  BSFP bsf;
  PTP pt[3];
  int i;
  FOR (i, 3) {
    if (ptid[i] == 0) {
      vdt_set_last_error_msg (vdt, "Point id <= 0");
      return FALSE;
    }
    pt[i] = ptid[i];
  }
  if (surf_id == 0) {
    vdt_set_last_error_msg (vdt, "Surface id <= 0");
    return FALSE;
  }
  if (reg_id[0] == VDT_INVALID_REG_ID) {
    vdt_set_last_error_msg (vdt, "Region id[0] not valid");
    return FALSE;
  }
  if (!vdt) {
    vdt_set_last_error_msg (vdt, "NULL passed for mesher");
    return FALSE;
  }
  bsf = vdt_bsf_new (pt, surf_id, reg_id);
  if (!bsf) {
    vdt_set_last_error_msg (vdt, "Failed to allocate new facet");
    return FALSE;
  }
  add_to_head (vdt->bsf_as_read, bsf);
  return TRUE;
}


/*
   Triangulate the interior.
 */
int /* FALSE if the triangulation failed for some reason */
VDT_triangulate (VDT vdt        /* mesher handle */
                 )
{
  if (!load_mesher (vdt)) return FALSE;
  return vdt_triangulate (vdt);
}

/*
   Produce the triangulation summary.
 */
void
VDT_summary (VDT vdt        /* mesher handle */
                 )
{
  vdt_summary (vdt);
}

/*
   Delete mesher.
 */
int /* TRUE if ok */
VDT_delete_mesher (VDT vdt /* mesher handle */
                   )
{
  return vdt_delete (vdt);
}


/*
   Get next tetrahedron.
   Pass prev=NULL to get the first; then pass the previously
   returned value to get the next.
 */
void *
VDT_next_tet (void *tetptid[4], /* identifiers of the tet points */
              void **reg_id, /* region identifier */
              void *adjtet[4], /* adjacent tets; opaque pointers as returned
                                  by this function
                                */
              void *prev, /* previous tet; opaque pointers as returned
                                  by this function.  Pass NULL to get the
                                  first tet.
                                  */
              VDT vdt /* mesher handle */
              )
{
  return vdt_next_tet (tetptid, reg_id, adjtet, prev, vdt);
}

 
/*
   Get next point.
   Pass prev=NULL to get the first; then pass the previously
   returned value to get the next.
 */
void *
VDT_next_pt (void **id, /* the identifier of the point; a unique number >0 */
             double *x, /* x-coordinate */
             double *y, /* y-coordinate */
             double *z, /* z-coordinate */
             double *mesh_size, /* mesh size */
             void *prev,/* previous point; opaque pointers as returned
                                  by this function.  Pass NULL to get the
                                  first point.
                                  */
             VDT vdt            /*  mesher handle  */
             ) 
{
  return vdt_next_pt (id, x, y, z, mesh_size, prev, vdt);
}


/*
   Get the next boundary surface facet.
   Pass prev=NULL to get the first; then pass the previously
   returned value to get the next.
 */
void *
VDT_next_bsf (void **id, /* the identifier of the facet */
              void *ptid[3],    /* identifiers of the points */
              void **surf_id,   /* identifier of the surface */
              void *reg_id[2], /* identifiers of regions on each side
                              of the facet */
              void *tet[2],     /* pointers to tetrahedra on each side
                              of the facet as returned by VDT_next_tet(); 
                              Note that tet[1] is on the
                              side to which the normal points,
                              tet[0] is on the other side.  */
              void *prev,/* previous boundary facet; opaque pointers as returned
                            by this function.  Pass NULL to get the
                            first facet
                            */
              VDT vdt        /*  mesher handle  */
              )
{
  return vdt_next_bsf (id, ptid, surf_id, reg_id, tet, prev, vdt);
}
     

/*
   Set option.
   The option is passed by setting the option number
   (as defined in the vdtpub.h header) and the option value.
 */
int /* TRUE if successful; FALSE otherwise */
VDT_set_option (VDT vdt,         /* mesher handle */
                vdt_option_t *option /* option */
                )
{
  return vdt_set_option (vdt, option);
}

/*
   Get default mesh size.
 */
double
VDT_def_mesh_size (VDT vdt) 
{
  vdt_t *vdtp = vdt;
  return vdtp->def_mesh_size;
}


/*
   Log module ids.
 */
void
VDT_log_module_ids (void)
{
  char buf[512];
  const char *vdt_vdtpub_module_id (void);
   
  logit ("Module ID's");

  sprintf (buf, "%s", VDT_MODULE_ID (vdt));       logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtaf));     logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtbsf));    logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtbucket)); logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtff));     logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtlp));     logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtoctree)); logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtpatch));  logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtpt));     logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtpub));    logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdttet));    logit (buf);
  sprintf (buf, "%s", VDT_MODULE_ID (vdtutil));   logit (buf);
}

/*
   Count points (both connected by tetrahedra and those unconnected).
 */
int
VDT_count_points (VDT vdt, int *connected, int *unconnected)
{
  return vdt_count_points (vdt, connected, unconnected);
}

/*
   Count input points.
 */
int
VDT_count_input_points (VDT vdt)
{
  return vdt->num_input_pts;
}

/*
   Count input boundary facets.
 */
int
VDT_count_input_bsfs (VDT vdt)
{
  return vdt->num_input_bsfs;
}

/*
   Count boundary facets.
 */
int
VDT_count_bsfs (VDT vdt)
{
  return list_length (vdt->bsfl);
}

/*
   Count tetrahedra.
 */
int
VDT_count_tets (VDT vdt)
{
  return vdt_count_tets (vdt);
}

/*
   Access message pertaining to last error raised by
   the public VDT interface.
 */
const char *
VDT_last_error_msg (VDT vdt)
{
  return (const char *) (vdt->last_error_msg ? vdt->last_error_msg : "?");
}

/*
   Get mesh quality.
 */
int
VDT_get_quality (VDT vdt,
                 int *num_non_positive_vol_tets,
                 int *num_small_dihed_angle_tets,
                 int *quality_measure,
                 double *min_quality,
                 double *max_quality,
                 double *geom_mean_quality,
                 double *rms_quality,
                 double *min_volume,
                 double *max_volume,
                 double *min_dihed_angle,
                 double *max_dihed_angle,
                 double *avg_dihed_angle
                 )
{
  return vdt_get_quality (vdt,
                          num_non_positive_vol_tets,
                          num_small_dihed_angle_tets,
                          quality_measure,
                          min_quality,
                          max_quality,
                          geom_mean_quality,
                          rms_quality,
                          min_volume,
                          max_volume,
                          min_dihed_angle,
                          max_dihed_angle,
                          avg_dihed_angle);
}

/*
   Compute tet dihedral angle quality.
 */
double
VDT_tet_get_angle_quality (void *obj,
                           double *min_angle,
                           double *max_angle,
                           double angles[6]
                           )
{
  return vdt_tet_angle_quality ((TETP)obj,
                                min_angle, max_angle, angles);
}

/*
   Get tet circumcenter.
 */
int /* TRUE if successful; FALSE otherwise, in which case
       the data in *cx, ..., *cz are not valid.
     */
VDT_tet_get_circumcenter (void *obj, VDT vdt,
                          double *cx, double *cy, double *cz)
{
  TETP tet = (TETP) obj;
  CkitPoint3 center={0,0,0};
  if (vdt_4_pt_circumsphere (tet->pt[0], tet->pt[1], tet->pt[2], tet->pt[3],
                             &center)) {
    *cx = center.x;
    *cy = center.y;
    *cz = center.z;
    return TRUE;
  } else {
    vdt_set_last_error_msg (vdt, "Failed to evaluate circumcenter");
    return FALSE;
  }
}

int
VDT_dump (VDT vdt)
{
  return vdt_dump (vdt);
}

#if GIFACE
void
VDT_generate_graphics (VDT vdt)
{
  vdt_generate_graphics (vdt);
}
#endif
